#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "statements.h"
#include "keywords.h"

float immed_fixpoint(char *fixpointval)
{
  int i=findpoint(fixpointval);
  char decimalpart[50];
  fixpointval[i]='\0';
  sprintf(decimalpart,"0.%s",fixpointval+i+1);
  return atof(decimalpart);
}

int findpoint(char *item)
{
  int i;
  for (i=0;i<50;++i) if (item[i] == '.') return i;
  return i;
}

void freemem(char **statement)
{
  int i;
  for (i=0;i<50;++i) free(statement[i]);
  free(statement);
}

void printfrac(char *item)
{ // prints the fractional part of a declared 8.8 fixpoint variable
  char getvar[50];
  int i;
  for (i=0;i<numfixpoint88;++i)
  {
    strcpy(getvar,fixpoint88[1][i]);
    if (!strcmp(fixpoint88[0][i],item)) 
    {
      printf("%s\n",getvar);
      return;
    }
  }
  // must be immediate value
  if (findpoint(item) < 50)
    printf("#%d\n",(int)(immed_fixpoint(item)*256.0));
  else printf("#0\n");
}

int isfixpoint(char *item)
{
  // determines if a variable is fixed point, and if so, what kind
  int i;
  removeCR(item);
  for (i=0;i<numfixpoint88;++i)
    if (!strcmp(item, fixpoint88[0][i])) return 8;
  for (i=0;i<numfixpoint44;++i)
    if (!strcmp(item, fixpoint44[0][i])) return 4;
  if (findpoint(item) < 50) return 12;
  return 0;
}

void set_romsize(char *size)
{
  if (!strncmp(size,"2k\0",2))
  {
    strcpy(redefined_variables[numredefvars++],"ROM2k = 1");
  }
  else if (strncmp(size,"4k\0",2))
  {
    fprintf(stderr,"Warning: unsupported ROM size: %s. Using 4k.\n",size);
  }
}

void add_includes(char *myinclude)
{
  strcat(user_includes,myinclude);
}

void init_includes()
{
  user_includes[0]='\0';
}

void create_includes(char *includesfile)
{
  FILE *includesread,*includeswrite;
  char line[500];
  int i;
  int writeline;
  if ((includesread = fopen(includesfile, "r")) == NULL)  // open file
  {  
    fprintf(stderr,"Cannot open %s for reading\n",includesfile);
    exit(1);
  }  
  if ((includeswrite = fopen("includes.bB", "w")) == NULL)  // open file
  {  
    fprintf(stderr,"Cannot open includes.bB for writing\n");
    exit(1);
  }  
  
  while (fgets(line,500,includesread))
  {
    for (i=0;i<500;++i)
    {
      if (line[i]==';') 
      {
	writeline=0;
	break;
      }
      if (line[i]==(unsigned char)0x0A) 
      {
	writeline=0;
	break;
      }
      if (line[i]==(unsigned char)0x0D) 
      {
	writeline=0;
	break;
      }
      if (line[i]>(unsigned char)0x20) 
      {
	writeline=1;
	break;
      }
      writeline=0;
    }
    if (writeline) 
    {
      if (!strncasecmp(line,"bb.asm\0",6)) 
	if (user_includes[0] != '\0') fprintf(includeswrite,user_includes);
      fprintf(includeswrite,line);
    }
  }
  fclose(includesread);
  fclose(includeswrite);
}

void printindex(char *mystatement, int myindex)
{
  if (!myindex)
  {
    printimmed(mystatement);
    printf("%s\n",mystatement);
  }
  else printf("%s,x\n",mystatement); // indexed with x!
}

void loadindex(char *myindex)
{
  printf("	LDX "); // get index
  printimmed(myindex);
  printf("%s\n",myindex);
}

int getindex(char *mystatement, char *myindex)
{
  int i,j,index=0;
  for (i=0;i<50;++i) if (mystatement[i]=='[') {index=1;break;}
  if (i<50)
  {
    strcpy(myindex,mystatement+i+1);
    myindex[strlen(myindex)-1]='\0';
    if (myindex[strlen(myindex)-2]==']')
      myindex[strlen(myindex)-2]='\0';
    if (myindex[strlen(myindex)-1]==']')
      myindex[strlen(myindex)-1]='\0';
    for (j=i;j<50;++j) mystatement[j]='\0';
  }
  return index;
}

int checkmul(int value)
{
// check to see if value can be optimized to save cycles

  if (value < 11) return 1;  // always optimize these

  while (value!=1)
  {
    if (!(value%9)) value/=9;
    else if (!(value%7)) value/=7;
    else if (!(value%5)) value/=5;
    else if (!(value%3)) value/=3;
    else if (!(value%2)) value/=2;
    else break;
    if (!(value%2) && optimization != 1) break; // do not optimize multplications
  }
  if (value == 1) return 1; else return 0;
}

int checkdiv(int value)
{
// check to see if value is a power of two - if not, run standard div routine
  while (value!=1)
  {
    if (value%2) break;
    else value/=2;
  }
  if (value == 1) return 1; else return 0;
}


void mul(char **statement, int bits)
{
  // this will attempt to output optimized code depending on the multiplicand
  int multiplicand=atoi(statement[6]);
  int tempstorage=0;
  // we will optimize specifically for 2,3,5,7,9
  if (bits == 16)
  {
    printf("	ldx #0\n");
    printf("	stx temp1\n");
  }
  while (multiplicand != 1)
  {
    if (!(multiplicand % 9))
    {
      if (tempstorage) 
      {
	strcpy(statement[4],"temp2");
	printf("	sta temp2\n");
      }
      multiplicand /= 9;
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	clc\n");
      printf("	adc ");
      printimmed(statement[4]);
      printf("%s\n",statement[4]);
      if (bits == 16)
      {
        printf("	tax\n");
        printf("	lda temp1\n");
        printf("	adc #0\n");
        printf("	sta temp1\n");
        printf("	txa\n");
      }
      tempstorage=1;
    }
    else if (!(multiplicand % 7))
    {
      if (tempstorage) 
      {
	strcpy(statement[4],"temp2");
	printf("	sta temp2\n");
      }
      multiplicand /= 7;
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	sec\n");
      printf("	sbc ");
      printimmed(statement[4]);
      printf("%s\n",statement[4]);
      if (bits == 16)
      {
        printf("	tax\n");
        printf("	lda temp1\n");
        printf("	sbc #0\n");
        printf("	sta temp1\n");
        printf("	txa\n");
      }
      tempstorage=1;
    }
    else if (!(multiplicand % 5))
    {
      if (tempstorage) 
      {
	strcpy(statement[4],"temp2");
	printf("	sta temp2\n");
      }
      multiplicand /= 5;
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	clc\n");
      printf("	adc ");
      printimmed(statement[4]);
      printf("%s\n",statement[4]);
      if (bits == 16)
      {
        printf("	tax\n");
        printf("	lda temp1\n");
        printf("	adc #0\n");
        printf("	sta temp1\n");
        printf("	txa\n");
      }
      tempstorage=1;
    }
    else if (!(multiplicand % 3))
    {
      if (tempstorage) 
      {
	strcpy(statement[4],"temp2");
	printf("	sta temp2\n");
      }
      multiplicand /= 3;
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
      printf("	clc\n");
      printf("	adc ");
      printimmed(statement[4]);
      printf("%s\n",statement[4]);
      if (bits == 16)
      {
        printf("	tax\n");
        printf("	lda temp1\n");
        printf("	adc #0\n");
        printf("	sta temp1\n");
        printf("	txa\n");
      }
      tempstorage=1;
    }
    else if (!(multiplicand % 2))
    {
      multiplicand /= 2;
      printf("	asl\n");
      if (bits == 16) printf("  rol temp1\n");
    }
    else 
    {
      printf("	LDY #%d\n",multiplicand);
      printf("	jsr mul%d\n",bits);
      fprintf(stderr,"Warning - there seems to be a problem.  Your code may not run properly.\n");
      fprintf(stderr,"If you are seeing this message, please report it - it could be a bug.\n");
    }
  }
}

void divd(char **statement, int bits)
{
  int divisor=atoi(statement[6]);
  if (bits == 16)
  {
    printf("	ldx #0\n");
    printf("	stx temp1\n");
  }
  while (divisor != 1)
  {
    if (!(divisor % 2)) // div by power of two is the only easy thing
    {
      divisor /= 2;
      printf("	lsr\n");
      if (bits == 16) printf("  rol temp1\n");  // I am not sure if this is actually correct
    }
    else 
    {
      printf("	LDY #%d\n",divisor);
      printf("	jsr div%d\n",bits);
      fprintf(stderr,"Warning - there seems to be a problem.  Your code may not run properly.\n");
      fprintf(stderr,"If you are seeing this message, please report it - it could be a bug.\n");
    }
  }
  
}

void endfunction()
{
  if (!doingfunction) prerror("extraneous end keyword encountered\n");
  doingfunction=0;
}

void function(char **statement)
{
  // declares a function - only needed if function is in bB.  For asm functions, see
  // the help.html file.
  // determine number of args, then run until we get an end. 
  doingfunction=1;
  printf("%s\n",statement[2]);
  printf("	STA temp1\n");
  printf("	STY temp2\n");
}

void callfunction(char **statement)
{
  // called by assignment to a variable
  // does not work as an argument in another function or an if-then... yet.
  int i,j,index=0;
  char getindex0[50];
  int arguments=0;
  int argnum[10];
  for (i=6;statement[i][0]!=')';++i)
  {
    if (statement[i][0] != ',') 
    {
      argnum[arguments++]=i;
    }
    if (i > 45) prerror("missing \")\" at end of function call\n");
  }
  if (!arguments) fprintf(stderr,"(%d) Warning: function call with no arguments\n",line);
  while (arguments)
  {
    // get [index]
    index=0;
    index |= getindex(statement[argnum[--arguments]],&getindex0[0]);
    if (index) loadindex(&getindex0[0]);

    if (arguments == 1) printf("	LDY ");
    else printf("	LDA ");
    printindex(statement[argnum[arguments]],index);

    if (arguments>1) printf("	STA temp%d\n",arguments+1);
//    arguments--;
  }
  printf("	jsr %s\n",statement[4]);
  strcpy(Areg,"invalid");

}

void incline()
{
  line++;
}

void invalidate_Areg()
{
  strcpy(Areg,"invalid");
}

int islabel(char **statement)
{ // this is for determining if the item after a "then" is a label or a statement
  // return of 0=label, 1=statement
  int i;
  // find the "then" or a "goto"
  for (i=0;i<48;) if (!strncmp(statement[i++],"then\0",4)) break;
  return findlabel(statement,i);
}

int islabelelse(char **statement)
{ // this is for determining if the item after an "else" is a label or a statement
  // return of 0=label, 1=statement
  int i;
  // find the "else"
  for (i=0;i<48;) if (!strncmp(statement[i++],"else\0",4)) break;
  return findlabel(statement,i);
}

int findlabel(char **statement, int i)
{
  if ((statement[i][0]>(unsigned char)0x2F) && (statement[i][0]<(unsigned char)0x3B)) return 0;
  if ((statement[i+1][0]==':') && (strncmp(statement[i+2],"rem\0",3))) return 1;
  if (!strncmp(statement[i+1],"else\0",4)) return 0;
  if (statement[i+1][0]!='\0') return 1;
  // only possibilities left are: drawscreen, player0:, player1:, asm, next, return, maybe others added later?
  if (!strncmp(statement[i],"drawscreen\0",10)) return 1;
  if (!strncmp(statement[i],"player0:\0",8)) return 1;
  if (!strncmp(statement[i],"player1:\0",8)) return 1;
  if (!strncmp(statement[i],"asm\0",3)) return 1;
  if (!strncmp(statement[i],"rem\0",3)) return 1;
  if (!strncmp(statement[i],"next\0",4)) return 1;
  if (!strncmp(statement[i],"return\0",6)) return 1;
  return 0; // I really hope we've got a label !!!!
}

void data(char **statement)
{
  char data[100];
  char **data_length;
  char **deallocdata_length;
  int i;
  data_length=(char **)malloc(sizeof(char *)*50);
  for (i=0;i<50;++i) data_length[i]=(char*)malloc(sizeof(char)*50);
  deallocdata_length=data_length;
  removeCR(statement[2]);
  printf("	JMP .skip%s\n",statement[0]);
  printf("%s\n",statement[2]);
  while (1) {
    if ( ((!fgets(data,100,stdin))
       || ( (data[0]<(unsigned char)0x3A) && (data[0]>(unsigned char)0x2F) )) 
          && (data[0] != 'e')) {
      prerror("Error: Missing \"end\" keyword at end of data\n");
      exit(1);
    }
    line++;
    if (!strncmp(data,"end\0",3)) break;
    remove_trailing_commas(data);
    printf("	.byte %s\n",data);
  }
  printf(".skip%s\n",statement[0]);
  strcpy(data_length[0]," ");
  strcpy(data_length[1],"const");
  sprintf(data_length[2],"%s_length",statement[2]);
  strcpy(data_length[3],"=");
  sprintf(data_length[4],".skip%s-%s",statement[0],statement[2]);
  strcpy(data_length[5],"\n");
  data_length[6][0]='\0';
  keywords(data_length);
  freemem(deallocdata_length);
}

void shiftdata(char **statement, int num)
{
  int i,j;
  for (i=49;i>num;i--)
    for (j=0;j<50;++j)
      statement[i][j]=statement[i-1][j];
}

void compressdata(char **statement, int num1, int num2)
{
  int i,j;
  for (i=num1;i<50-num2;i++)
    for (j=0;j<50;++j)
      statement[i][j]=statement[i+num2][j];
}

void ongoto(char **statement)
{
  int k,i=4;
  if (strcmp(statement[2],Areg)) printf("	LDA %s\n",statement[2]);
  printf("	ASL\n");
  printf("	TAX\n");
  printf("	LDA .%sjumptable,x\n",statement[0]);
  printf("	PHA\n");
  printf("	INX\n");
  printf("	LDA .%sjumptable,x\n",statement[0]);
  printf("	PHA\n");
  printf("	RTS\n");
  printf(".%sjumptable\n",statement[0]);
  while ((statement[i][0]!=':') && (statement[i][0]!='\0'))
  {
    for (k=0;k<100;++k)
      if ((statement[i][k]==(unsigned char)0x0A) || (statement[i][k]==(unsigned char)0x0D))
	statement[i][k]='\0';
    printf("	.byte >(.%s-1)\n",statement[i]);
    printf("	.byte <(.%s-1)\n",statement[i++]);
  }
}

void dofor(char **statement)
{
  if (strcmp(statement[4],Areg)) 
  {
    printf("	LDA ");
    printimmed(statement[4]);
    printf("%s\n",statement[4]);
  }
  
  printf("	STA %s\n",statement[2]);
  
  forlabel[numfors][0]='\0';
  sprintf(forlabel[numfors],"%sfor%s",statement[0],statement[2]);
  printf(".%s\n",forlabel[numfors]);

  forend[numfors][0]='\0';
  strcpy(forend[numfors],statement[6]);

  forvar[numfors][0]='\0';
  strcpy(forvar[numfors],statement[2]);

  forstep[numfors][0]='\0';

  if (!strncasecmp(statement[7],"step\0",4)) strcpy(forstep[numfors],statement[8]);
  else strcpy(forstep[numfors],"1");

  numfors++;
}

void next(char **statement)
{
  int immed=0;
  int immedend=0;
  int failsafe=0;
  char failsafelabel[60];

  invalidate_Areg();

  if (!numfors) {
    fprintf(stderr,"(%d) next without for\n",line);
    exit(1);
  }
  numfors--;
  if (!strncmp(forstep[numfors],"1\0",2)) // step 1
  {
    printf("	LDA %s\n",forvar[numfors]);
    printf("	CMP ");
    printimmed(forend[numfors]);
    printf("%s\n",forend[numfors]);
    printf("	INC %s\n",forvar[numfors]);
    bcc(forlabel[numfors]);
  }
  else if ((!strncmp(forstep[numfors],"-1\0",3)) || (!strncmp(forstep[numfors],"255\0",4)))
  { // step -1
    printf("	DEC %s\n",forvar[numfors]);
    if (strncmp(forend[numfors],"1\0",2))
    {
      printf("	LDA %s\n",forvar[numfors]);
      printf("	CMP ");
      printimmed(forend[numfors]);
      printf("%s\n",forend[numfors]);
      bcs(forlabel[numfors]);    
    }
    else
      bne(forlabel[numfors]);    
  }
  else // step other than 1 or -1
  {
    // normally, the generated code adds to or subtracts from the for variable, and checks
    // to see if it's less than the ending value.
    // however, if the step would make the variable less than zero or more than 255 
    // then this check will not work.  The compiler will attempt to detect this situation
    // if the step and end are known.  If the step and end are not known (that is, 
    // either is a variable) then much more complex code must be generated.

    printf("	LDA %s\n",forvar[numfors]);
    printf("	CLC\n");
    printf("	ADC ");
    immed=printimmed(forstep[numfors]);
    printf("%s\n",forstep[numfors]);

    if (immed && isimmed(forend[numfors])) // the step and end are immediate
    {
      if (atoi(forstep[numfors])&128) // step is negative
      {
	if ((256-(atoi(forstep[numfors])&255)) >= atoi(forend[numfors]))
	{ // if we are in danger of going < 0...we will have carry clear after ADC
	  failsafe=1;
	  sprintf(failsafelabel,"%s_failsafe",forlabel[numfors]);
	  bcc(failsafelabel);
	}
      }
      else { // step is positive
	if ((atoi(forstep[numfors]) + atoi(forend[numfors])) > 255)
	{ // if we are in danger of going > 255...we will have carry set after ADC
	  failsafe=1;
	  sprintf(failsafelabel,"%s_failsafe",forlabel[numfors]);
	  bcs(failsafelabel);
	}
      }

    }
    printf("	STA %s\n",forvar[numfors]);

    printf("	CMP ");
    immedend=printimmed(forend[numfors]);
    // add 1 to immediate compare for increasing loops
    if (immedend && !(atoi(forstep[numfors])&128))
      strcat(forend[numfors],"+1");
    printf("%s\n",forend[numfors]);
// if step number is 1 to 127 then add 1 and use bcc, otherwise bcs
// if step is a variable, we'll need to check for every loop iteration
//
// Warning! no failsafe checks with variables as step or end - it's the
// programmer's job to make sure the end value doesn't overflow
    if (!immed) 
    {
      printf("	LDX %s\n",forstep[numfors]);
      printf("	BMI .%sbcs\n",statement[0]);
      bcc(forlabel[numfors]);
      printf("	CLC\n");
      printf(".%sbcs\n",statement[0]);
      bcs(forlabel[numfors]);
    }
    else if (atoi(forstep[numfors])&128) bcs(forlabel[numfors]);
    else
    {
      bcc(forlabel[numfors]);
      if (!immedend) beq(forlabel[numfors]);
    }
  }
  if (failsafe) printf(".%s\n",failsafelabel);
}

void dim(char **statement)
{
  // just take the statement and pass it right to a header file
  int i;
  char *fixpointvar1;
  char fixpointvar2[50];
  // check for fixedpoint variables
  i=findpoint(statement[4]);
  if (i<50)
  {
    removeCR(statement[4]);
    strcpy(fixpointvar2,statement[4]);
    fixpointvar1=fixpointvar2+i+1; 
    fixpointvar2[i]='\0'; 
    if (!strcmp(fixpointvar1,fixpointvar2)) // we have 4.4
      {
	strcpy(fixpoint44[1][numfixpoint44],fixpointvar1);
	strcpy(fixpoint44[0][numfixpoint44++],statement[2]);
      }
    else // we have 8.8
      {
	strcpy(fixpoint88[1][numfixpoint88],fixpointvar1);
	strcpy(fixpoint88[0][numfixpoint88++],statement[2]);
      }
    statement[4][i]='\0'; // terminate string at '.'
  }
  i=2;
  redefined_variables[numredefvars][0]='\0';
  while (statement[i][0] != '\0') 
  {
    strcat(redefined_variables[numredefvars],statement[i++]);
    strcat(redefined_variables[numredefvars]," ");
  }
  numredefvars++;
}

void doconst(char **statement)
{
  // basically the same as dim, except we keep a queue of variable names that are constant
  int i=2;
  redefined_variables[numredefvars][0]='\0';
  while (statement[i][0] != '\0')
  {
    strcat(redefined_variables[numredefvars],statement[i++]);
    strcat(redefined_variables[numredefvars]," ");
  }
  numredefvars++;
  strcpy(constants[numconstants++],statement[2]); // record to queue
}

void doreturn(char **statement)
{
  int i,j,index=0;
  char getindex0[50];
  //if (!numgosubs) {
    //fprintf(stderr,"return without gosub in line %s\n",statement[0]);
    //exit(1);
  //}
  //numgosubs--;
  if (statement[2][0] && (statement[2][0]!=' '))
  {

    index |= getindex(statement[2],&getindex0[0]);

    if (index&1) loadindex(&getindex0[0]);

    printf("	LDA ");
    printindex(statement[2],index&1);
    
  }
  printf("	RTS\n");
}

void pfread(char **statement)
{
  char getindex0[50];
  char getindex1[50];
  int i,j,index=0;

  invalidate_Areg();

  index |= getindex(statement[3],&getindex0[0]);
  index |= getindex(statement[4],&getindex1[0])<<1;

  if (index&1) loadindex(&getindex0[0]);

  printf("	LDA ");
  printindex(statement[4],index&1);

  if (index&2) loadindex(&getindex1[0]);

  printf("	LDY ");
  printindex(statement[6],index&2);

  printf("	jsr pfread\n");
}

void pfpixel(char **statement)
{
  char getindex0[50];
  char getindex1[50];
  int i,j,index=0;

  invalidate_Areg();

  index |= getindex(statement[2],&getindex0[0]);
  index |= getindex(statement[3],&getindex1[0])<<1;

  if (index&1) loadindex(&getindex0[0]);

  printf("	LDA ");
  printindex(statement[2],index&1);

  if (index&2) loadindex(&getindex1[0]);

  printf("	LDY ");
  printindex(statement[3],index&2);

  printf("	LDX #");
  if (!strncmp(statement[4],"flip",2)) printf("2\n");
  else if (!strncmp(statement[4],"off",2)) printf("1\n");
  else printf("0\n");
  printf("	jsr pfpixel\n");
}

void pfhline(char **statement)
{
  char getindex0[50];
  char getindex1[50];
  char getindex2[50];
  int i,j,index=0;

  invalidate_Areg();

  index |= getindex(statement[2],&getindex0[0]);
  index |= getindex(statement[3],&getindex1[0])<<1;
  index |= getindex(statement[4],&getindex2[0])<<2;

  if (index&4) loadindex(&getindex2[0]);
  printf("	LDA ");
  printindex(statement[4],index&4);
  printf("	STA temp3\n");

  if (index&1) loadindex(&getindex0[0]);
  printf("	LDA ");
  printindex(statement[2],index&1);

  if (index&2) loadindex(&getindex1[0]);
  printf("	LDY ");
  printindex(statement[3],index&2);

  printf("	LDX #");
  if (!strncmp(statement[5],"flip",2)) printf("2\n");
  else if (!strncmp(statement[5],"off",2)) printf("1\n");
  else printf("0\n");

  printf("	jsr pfhline\n");
}

void pfvline(char **statement)
{
  char getindex0[50];
  char getindex1[50];
  char getindex2[50];
  int i,j,index=0;

  invalidate_Areg();

  index |= getindex(statement[2],&getindex0[0]);
  index |= getindex(statement[3],&getindex1[0])<<1;
  index |= getindex(statement[4],&getindex2[0])<<2;

  if (index&4) loadindex(&getindex2[0]);
  printf("	LDA ");
  printindex(statement[4],index&4);
  printf("	STA temp3\n");

  if (index&1) loadindex(&getindex0[0]);
  printf("	LDA ");
  printindex(statement[2],index&1);

  if (index&2) loadindex(&getindex1[0]);
  printf("	LDY ");
  printindex(statement[3],index&2);

  printf("	LDX #");
  if (!strncmp(statement[5],"flip",2)) printf("2\n");
  else if (!strncmp(statement[5],"off",2)) printf("1\n");
  else printf("0\n");

  printf("	jsr pfvline\n");
}

void pfscroll(char **statement)
{
  invalidate_Areg();

  printf("	LDA #");
  if (!strncmp(statement[2],"left",2)) printf("0\n");
  else if (!strncmp(statement[2],"right",2)) printf("1\n");
  else if (!strncmp(statement[2],"up\0",2)) printf("2\n");
  else if (!strncmp(statement[2],"down",2)) printf("4\n");
  else {fprintf(stderr,"(%d) pfscroll direction unknown\n",line);exit(1);}

  printf("	jsr pfscroll\n");
}

void doasm()
{
  char data[100];
  while (1) {
    if ( ((!fgets(data,100,stdin))
       || ( (data[0]<(unsigned char)0x3A) && (data[0]>(unsigned char)0x2F) )) 
          && (data[0] != 'e')) {
      prerror("Error: Missing \"end\" keyword at end of inline asm\n");
      exit(1);
    }
    line++;
    if (!strncmp(data,"end\0",3)) break;
    printf("%s\n",data);  

  }
}

void player(char **statement)
{
  int height=0,i=0; //calculating sprite height
  char label[20];
  char data[100];
  char bytes[1000];
  char byte[100];
  char pl=statement[1][6];
  bytes[0]='\0';
  sprintf(label,"player%s_%c\n",statement[0],pl);
  
  printf("	LDA #<%s\n",label);
  printf("	STA player%cpointer\n",pl);

  printf("	LDA #>%s\n",label);
  printf("	STA player%cpointer+1\n",pl);


  printf("	JMP .%sjump%c\n",statement[0],pl);
 
  while (1) {
    if ( ((!fgets(data,100,stdin))
       || ( (data[0]<(unsigned char)0x3A) && (data[0]>(unsigned char)0x2F) )) 
          && (data[0] != 'e')) {

      prerror("Error: Missing \"end\" keyword at end of player declaration\n");
      exit(1);
    }
    line++;
    if (!strncmp(data,"end\0",3)) break;
    height++;
    sprintf(byte,"	.byte %s",data);  
    strcat(bytes,byte);
 //   printf("	.byte %s\n",data);  
 
  }

  // insert DASM stuff to prevent page-wrapping of player data
  printf(" if (<*) > (<(*+%d))\n	repeat ($100-<*)\n	nop\n	repend\n	endif\n",height);

  printf("%s\n",label);
  printf("	.byte 0\n");

  printf("%s\n",bytes);

  printf(".%sjump%c\n",statement[0],pl);
//  while (statement[2][i]!='\0') {

//    if (statement[2][i++]==',') height++;
//  }  
  printf("	LDA #%d\n",height);
  printf("	STA player%cheight\n",pl);
}

void doif(char **statement)
{
  int index=0;
  char getindex0[50];
  char getindex1[50];
  int not=0;
  int i,j,k;
  int bit=0;
  int Aregmatch=0;
  char Aregcopy[50];
  char **cstatement; //conditional statement
  char **dealloccstatement; //for deallocation
  
  strcpy(Aregcopy,"index-invalid");

  cstatement = (char **)malloc(sizeof(char*)*50);
  for (k=0;k<50;++k) cstatement[k]=(char*)malloc(sizeof(char)*50);
  dealloccstatement=cstatement;
  for (k=0;k<50;++k) for (j=0;j<50;++j) cstatement[j][k]='\0';
  if (statement[2][0] == '!') {
    not=1;
    for (i=0;i<49;++i) {
      statement[2][i]=statement[2][i+1];
    }
  }
  if ( (!strncmp(statement[2],"joy\0",3)) 
   || (!strncmp(statement[2],"switch\0",6)) ) {
    printf("	jsr %s\n",statement[2]);
    if (!islabel(statement)) {
      if (not) bne(statement[4]); else beq(statement[4]);
//printf("	BNE "); else printf("	BEQ ");
//      printf(".%s\n",statement[4]);
      freemem(dealloccstatement);
      return;
    }
    else // then statement
    {
      if (not) printf("	BEQ "); else printf("	BNE ");

      printf(".skip%s\n",statement[0]);
      // separate statement
      for (i=3;i<50;++i)
      {
        for (k=0;k<50;++k) {
          cstatement[i-3][k]=statement[i][k];
	}
      }
      keywords(cstatement);
      printf(".skip%s\n",statement[0]);
 
      freemem(dealloccstatement);
      return;      
    }
  }

  if (!strncmp(statement[2],"pfread\0",6)) { 
//    printf("	jsr pfread\n");
    pfread(statement);
    if (!islabel(statement)) {
      if (not) bne(statement[9]); else beq(statement[9]);
      freemem(dealloccstatement);
      return;
    }
    else // then statement
    {
      if (not) printf("	BEQ "); else printf("	BNE ");

      printf(".skip%s\n",statement[0]);
      // separate statement
      for (i=8;i<50;++i)
      {
        for (k=0;k<50;++k) {
          cstatement[i-8][k]=statement[i][k];
	}
      }
      keywords(cstatement);
      printf(".skip%s\n",statement[0]);
      freemem(dealloccstatement);
      return;      
    }
  }

  if  (!strncmp(statement[2],"collision(\0",10)) {
    if (!strncmp(statement[2],"collision(missile0,player1)\0",27)) {
      printf("	BIT CXM0P\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(missile0,player0)\0",27)) {
      printf("	BIT CXM0P\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(missile1,player0)\0",27)) {
      printf("	BIT CXM1P\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(missile1,player1)\0",27)) {
      printf("	BIT CXM1P\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(player0,playfield)\0",28)) {
      printf("	BIT CXP0FB\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(player0,ball)\0",23)) {
      printf("	BIT CXP0FB\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(player1,playfield)\0",28)) {
      printf("	BIT CXP1FB\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(player1,ball)\0",23)) {
      printf("	BIT CXP1FB\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(missile0,playfield)\0",29)) {
      printf("	BIT CXM0FB\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(missile0,ball)\0",24)) {
      printf("	BIT CXM0FB\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(missile1,playfield)\0",29)) {
      printf("	BIT CXM1FB\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(missile1,ball)\0",24)) {
      printf("	BIT CXM1FB\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(ball,playfield)\0",25)) {
      printf("	BIT CXBLPF\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(player0,player1)\0",26)) {
      printf("	BIT CXPPMM\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(missile0,missile1)\0",28)) {
      printf("	BIT CXPPMM\n");
      bit=6;
    } // now repeat everything in reverse...


    else if (!strncmp(statement[2],"collision(player1,missile0)\0",27)) {
      printf("	BIT CXM0P\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(player0,missile0)\0",27)) {
      printf("	BIT CXM0P\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(player0,missile1)\0",27)) {
      printf("	BIT CXM1P\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(player1,missile1)\0",27)) {
      printf("	BIT CXM1P\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(playfield,player0)\0",28)) {
      printf("	BIT CXP0FB\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(ball,player0)\0",23)) {
      printf("	BIT CXP0FB\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(playfield,player1)\0",28)) {
      printf("	BIT CXP1FB\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(ball,player1)\0",23)) {
      printf("	BIT CXP1FB\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(playfield,missile0)\0",29)) {
      printf("	BIT CXM0FB\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(ball,missile0)\0",24)) {
      printf("	BIT CXM0FB\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(playfield,missile1)\0",29)) {
      printf("	BIT CXM1FB\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(ball,missile1)\0",24)) {
      printf("	BIT CXM1FB\n");
      bit=6;
    }
    else if (!strncmp(statement[2],"collision(playfield,ball)\0",25)) {
      printf("	BIT CXBLPF\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(player1,player0)\0",26)) {
      printf("	BIT CXPPMM\n");
      bit=7;
    }
    else if (!strncmp(statement[2],"collision(missile1,missile0)\0",28)) {
      printf("	BIT CXPPMM\n");
      bit=6;
    }



    else //error
    {
      fprintf(stderr, "(%d) Error: cannot parse if-then statement\n",line);
      exit(1);
    }


    if (!islabel(statement)) {
      if (!not) {if (bit==7) bmi(statement[4]); else bvs(statement[4]);}
      else {if (bit==7) bpl(statement[4]); else bvc(statement[4]);}
//{if (bit==7) printf("	BMI "); else printf("	BVS ");}
//      else {if (bit==7) printf("	BPL "); else printf("	BVC ");}
//      printf(".%s\n",statement[4]);
      freemem(dealloccstatement);
      return;
    }
    else // then statement
    {
      if (not) {if (bit==7) printf("	BMI "); else printf("	BVS ");}
      else {if (bit==7) printf("	BPL "); else printf("	BVC ");}

      printf(".skip%s\n",statement[0]);
      // separate statement
      for (i=3;i<50;++i)
      {
        for (k=0;k<50;++k) {
          cstatement[i-3][k]=statement[i][k];
	}
      }
      keywords(cstatement);
      printf(".skip%s\n",statement[0]);
 
      freemem(dealloccstatement);
      return;      
    }
  }


  // check for array, e.g. x(1) to get bit 1 of x
  for (i=0;i<50;++i) if (statement[2][i]=='}') break;

  if (i<50) // found array
  {
    // extract expression in parantheses - for now just whole numbers allowed
    bit=(int)statement[2][i-1]-48;
    if ((bit>9)||(bit<0))
    {
      fprintf(stderr,"(%d) Error: variables in bit access not supported\n",line);
      exit(1);
    }
    if ((bit==7) || (bit==6)) // if bit 6 or 7, we can use BIT and save 2 bytes
    {
      printf("	BIT ");
      for (i=0;i<50;++i) 
      {
	if (statement[2][i]=='{') break;
	printf("%c",statement[2][i]);
      }
      printf("\n");
      if (!islabel(statement)) {
        if (!not) {if (bit==7) bmi(statement[4]); else bvs(statement[4]);}
        else {if (bit==7) bpl(statement[4]); else bvc(statement[4]);}
        freemem(dealloccstatement);
        return;
      }
      else // then statement
      {
        if (not) {if (bit==7) printf("	BMI "); else printf("	BVS ");}
        else {if (bit==7) printf("	BPL "); else printf("	BVC ");}

        printf(".skip%s\n",statement[0]);
        // separate statement
        for (i=3;i<50;++i)
        {
          for (k=0;k<50;++k) {
            cstatement[i-3][k]=statement[i][k];
  	  }
        }
        keywords(cstatement);
        printf(".skip%s\n",statement[0]);
 
        freemem(dealloccstatement);
        return;      
      }
    }
    else
    { 
      Aregmatch=0;
      printf("	LDA ");
      for (i=0;i<50;++i) 
      {
	if (statement[2][i]=='{') break;
	printf("%c",statement[2][i]);
      }
      printf("\n");
      if (!bit)      // if bit 0, we can use LSR and save a byte
	printf("	LSR\n");
      else printf("	AND #%d\n",1<<bit);//(int)pow(2,bit));
      if (!islabel(statement)) {
        if (!not) {if (!bit) bcc(statement[4]); else beq(statement[4]);}
        else {if (!bit) bcs(statement[4]); else bne(statement[4]);}
        freemem(dealloccstatement);
        return;
      }
      else // then statement
      {
        if (not) {if (!bit) printf("	BCS "); else printf("	BNE ");}
        else {if (!bit) printf("	BCC "); else printf("	BEQ ");}

        printf(".skip%s\n",statement[0]);
        // separate statement
        for (i=3;i<50;++i)
        {
          for (k=0;k<50;++k) {
            cstatement[i-3][k]=statement[i][k];
  	  }
        }
        keywords(cstatement);
        printf(".skip%s\n",statement[0]);
 
        freemem(dealloccstatement);
        return;      
      }



    }

  }
  
// generic if-then (no special considerations)
  //check for [indexing]
  strcpy(Aregcopy,statement[2]);
  if (!strcmp(statement[2],Areg)) Aregmatch=1; // do we already have the correct value in A?

  index |= getindex(statement[2],&getindex0[0]);
  if (strncmp(statement[3],"then\0",4))
    index |= getindex(statement[4],&getindex1[0])<<1;

    if (!Aregmatch) // do we already have the correct value in A?
    {
      if (index&1) loadindex(&getindex0[0]);
      printf("	LDA ");
      printindex(statement[2],index&1);
      strcpy(Areg,Aregcopy);
    }
    if (index&2) loadindex(&getindex1[0]);
  printf("	CMP ");
  printindex(statement[4],index&2);

  if (!islabel(statement)){ // then linenumber
    if (statement[3][0] == '=') beq(statement[6]);
    if (!strncmp(statement[3],"<>\0",2)) bne(statement[6]);
    else if (statement[3][0] == '<') bcc(statement[6]);
    if (statement[3][0] == '>') bcs(statement[6]);
//    if (statement[3][0] == '=') printf ("     BEQ ");
//    if (!strcmp(statement[3],"<>")) printf ("     BNE ");
//    else if (statement[3][0] == '<') printf ("     BCC ");
//    if (statement[3][0] == '>') printf ("     BCS ");
  
//    printf(".%s\n",statement[6]);
  }
  else // then statement
  { // first, take negative of condition and branch around statement
    if (statement[3][0] == '=') printf ("     BNE ");
    if (!strcmp(statement[3],"<>")) printf ("     BEQ ");
    else if (statement[3][0] == '<') printf ("     BCS ");
    if (statement[3][0] == '>') printf ("     BCC ");
  
    printf(".skip%s\n",statement[0]);
    // separate statement

      // separate statement
      for (i=5;i<50;++i)
      {
        for (k=0;k<50;++k) {
          cstatement[i-5][k]=statement[i][k];
	}
      }
      keywords(cstatement);
      printf(".skip%s\n",statement[0]);
 
      freemem(dealloccstatement);
      return;      


//    i=4;
//    while (statement[i][0]!='\0')
//    {
//      cstatement[i-4]=statement[i++];
//    }
//    keywords(cstatement);
//    printf(".skip%s\n",statement[0]);
  }
  freemem(dealloccstatement);
}

void let(char **cstatement)
{
  int i,j=0,bit=0;
  int index=0;
  int immediate=0;
  char getindex0[50];
  char getindex1[50];
  char getindex2[50];
  int score[6]={0,0,0,0,0,0};
  char **statement;
  char *getbitvar;
  int Aregmatch=0;
  char Aregcopy[50];
  int fixpoint1;
  int fixpoint2;
  int fixpoint3=0;
  char **deallocstatement;
  strcpy(Aregcopy,"index-invalid");

  statement = (char **)malloc(sizeof(char*)*50);
  deallocstatement=statement;
  if (!strncmp(cstatement[2],"=\0",1)) {
    for (i=10;i>0;--i) {
      statement[i+1]=cstatement[i];
    }
  }
  else statement=cstatement;

  fixpoint1=isfixpoint(statement[2]);
  fixpoint2=isfixpoint(statement[4]);

  //check for [indexing]
  removeCR(statement[4]);
  strcpy(Aregcopy,statement[4]);
  if (!strcmp(statement[4],Areg)) Aregmatch=1; // do we already have the correct value in A?

  index |= getindex(statement[2],&getindex0[0]);
  index |= getindex(statement[4],&getindex1[0])<<1;
  if (statement[5][0]!=':')
    index |= getindex(statement[6],&getindex2[0])<<2;

  // check for array, e.g. x(1) to access bit 1 of x
  for (i=0;i<50;++i) if (statement[2][i]=='}') break;
  if (i<50) // found bit
  {
    strcpy(Areg,"invalid");
    // extract expression in parantheses - for now just whole numbers allowed
    bit=(int)statement[2][i-1]-48;
    if ((bit>9)||(bit<0))
    {
      fprintf(stderr,"(%d) Error: variables in bit access not supported\n",line);
      exit(1);
    }

    if (statement[4][0]=='0')
    {
      printf("	LDA ");
      for (i=0;i<50;++i) 
      {
        if (statement[2][i]=='{') break;
        printf("%c",statement[2][i]);
      }
      printf("\n");
      printf("	AND #%d\n",255^(1<<bit));//(int)pow(2,bit));
    }
    else if (statement[4][0]=='1')
    {
      printf("	LDA ");
      for (i=0;i<50;++i) 
      {
        if (statement[2][i]=='{') break;
        printf("%c",statement[2][i]);
      }
      printf("\n");
      printf("	ORA #%d\n",1<<bit);//(int)pow(2,bit));
    }
    else if (getbitvar=strtok(statement[4],"{"))
    {  // assign one bit to another
	// I haven't a clue if this will actually work!
      if (getbitvar[0] == '!') 
	printf("	LDA %s\n",getbitvar+1);
      else
	printf("	LDA %s\n",getbitvar);
      printf("	AND #%d\n",(1<<atoi(statement[4])));
      printf("	PHP\n");

      printf("	LDA ");
      for (i=0;i<50;++i) 
      {
        if (statement[2][i]=='{') break;
        printf("%c",statement[2][i]);
      }
      printf("\n	PLP\n");
      if (getbitvar[0] == '!') 
        printf("	.byte $F0, $05\n"); //bad, bad way to do BEQ addr+5
      else
        printf("	.byte $D0, $05\n"); //bad, bad way to do BNE addr+5
      printf("\n");

      printf("	AND #%d\n",255^(1<<bit));//(int)pow(2,bit));
      printf("	.byte $2C\n"); // BIT absoulte to skip next two bytes
      printf("	ORA #%d\n",1<<bit);//(int)pow(2,bit));

    }
    else 
    {
      fprintf(stderr,"(%d) Error: can only assign 0, 1 or another bit to a bit\n",line);
      exit(1);
    }
    printf("	STA ");
    for (i=0;i<50;++i) 
    {
      if (statement[2][i]=='{') break;
      printf("%c",statement[2][i]);
    }
    printf("\n");
    free(deallocstatement);
    return;
  }

  if (statement[4][0]=='-') // assignment to negative
  {
    strcpy(Areg,"invalid");
    if ((!fixpoint1) && (isfixpoint(statement[5]) != 12))
    {
      if (statement[5][0]>(unsigned char)0x39) // perhaps include constants too?
      {
        printf("	LDA #0\n");
        printf("  SEC\n");
        printf("	SBC %s\n",statement[5]);
      }
      else printf("	LDA #%d\n",256-atoi(statement[5]));
    }
    else
    {
      if (fixpoint1 == 4)
      {
      if (statement[5][0]>(unsigned char)0x39) // perhaps include constants too?
	{
          printf("	LDA #0\n");
          printf("  SEC\n");
          printf("	SBC %s\n",statement[5]);
	}
	else
	  printf("	LDA #%d\n",(int)((16-atof(statement[5]))*16));
	printf("	STA %s\n",statement[2]);
	free(deallocstatement);
	return;
      }
      if (fixpoint1 == 8)
      {
	printf("	LDX ");
	sprintf(statement[4],"%f",256.0-atof(statement[5]));
	printfrac(statement[4]);
	printf("	STX ");
	printfrac(statement[2]);
	printf("	LDA #%s\n",statement[4]);
	printf("	STA %s\n",statement[2]);
	free(deallocstatement);
	return;
      }
    }
  }
  else if (!strncmp(statement[4],"rand\0",4)) {
    strcpy(Areg,"invalid");
    printf("	  JSR randomize\n");
  }
  else if ( (!strncmp(statement[2],"score\0",5))
            && (strncmp(statement[2],"scorec\0",6)))
  {
// break up into three parts
    strcpy(Areg,"invalid");
    if (statement[5][0] == '+')
    {
      printf("	SED\n");
      printf("	CLC\n");
      for (i=5;i>=0;i--)
      {
  	if (statement[6][i] != '\0') score[j]=number(statement[6][i]);
	score[j]=number(statement[6][i]);
	if ((score[j]<10) && (score[j]>=0)) j++;
      }
      printf("	LDA score+2\n");
      if (statement[6][0]>(unsigned char)0x3F) printf("	ADC %s\n",statement[6]);
      else printf("	ADC #$%d%d\n",score[1],score[0]);
      printf("	STA score+2\n");      
      printf("	LDA score+1\n");      
      if (score[0]>9) printf("	ADC #0\n");
      else printf("	ADC #$%d%d\n",score[3],score[2]);
      printf("	STA score+1\n");      
      printf("	LDA score\n");      
      if (score[0]>9) printf("	ADC #0\n");
      else printf("	ADC #$%d%d\n",score[5],score[4]);
      printf("	STA score\n");      
      printf("	CLD\n");
    }
    else if (statement[5][0] == '-')
    {
      printf("	SED\n");
      printf("	SEC\n");
      for (i=5;i>=0;i--)
      {
	if (statement[6][i] != '\0') score[j]=number(statement[6][i]);
	score[j]=number(statement[6][i]);
	if ((score[j]<10) && (score[j]>=0)) j++;
      }
      printf("	LDA score+2\n");      
      if (score[0]>9) printf("	SBC %s\n",statement[6]);
      else printf("	SBC #$%d%d\n",score[1],score[0]);
      printf("	STA score+2\n");      
      printf("	LDA score+1\n");      
      if (score[0]>9) printf("	SBC #0\n");
      else printf("	SBC #$%d%d\n",score[3],score[2]);
      printf("	STA score+1\n");      
      printf("	LDA score\n");      
      if (score[0]>9) printf("	SBC #0\n");
      else printf("	SBC #$%d%d\n",score[5],score[4]);
      printf("	STA score\n");      
      printf("	CLD\n");
    }
    else
    {
      for (i=5;i>=0;i--)
      {
	if (statement[4][i] != '\0') score[j]=number(statement[4][i]);
	score[j]=number(statement[4][i]);
	if ((score[j]<10) && (score[j]>=0)) j++;
      }
      printf("	LDA #$%d%d\n",score[1],score[0]);
      printf("	STA score+2\n");      
      printf("	LDA #$%d%d\n",score[3],score[2]);
      printf("	STA score+1\n");      
      printf("	LDA #$%d%d\n",score[5],score[4]);
      printf("	STA score\n");      
    }
    free(deallocstatement);
    return;

  }
  else if ( (statement[6][0] == '1') && ((statement[6][1] > (unsigned char)0x39) 
          || (statement[6][1] < (unsigned char)0x30)) &&
          ((statement[5][0] == '+') || (statement[5][0] == '-')) && 
	  (!strncmp(statement[2], statement[4], 50)) && 
          (statement[6][1]=='\0' || statement[6][1]==' '))
  { // var=var +/- something
    strcpy(Areg,"invalid");
    if ((fixpoint1 == 4) && (fixpoint2 == 4))
    {
      if (statement[5][0] == '+')
      {
	printf("	LDA %s\n",statement[2]);
	printf("	CLC %s\n");
	printf("	ADC #16\n");
	printf("	STA %s\n",statement[2]);
	free(deallocstatement);
	return;
      }
      if (statement[5][0] == '+')
      {
	printf("	LDA %s\n",statement[2]);
	printf("	SEC %s\n");
	printf("	SBC #16\n");
	printf("	STA %s\n",statement[2]);
	free(deallocstatement);
	return;
      }    
    }

    if (index&1) loadindex(&getindex0[0]);
    if (statement[5][0] == '+') printf("	INC ");
    else printf("	DEC ");
    if (!(index&1))
      printf("%s\n",statement[2]);
    else printf("%s,x\n",statement[4]); // indexed with x!
    free(deallocstatement);

    return;
  }
  else 
  {// This is generic x=num or var

    if (!Aregmatch) // do we already have the correct value in A?
    {
      if (index&2) loadindex(&getindex1[0]);

// if 8.8=8.8+8.8: this LDA will be superfluous - fix this at some point   

//      if (!fixpoint1 && !fixpoint2 && statement[5][0]!='(')
      if (((!fixpoint1 && !fixpoint2) || (!fixpoint1 && fixpoint2 == 8)) && statement[5][0]!='(')

//	printfrac(statement[4]);
//      else
      {
        printf("	LDA ");
	printindex(statement[4],index&2);
      }
      strcpy(Areg,Aregcopy);
    }
  }
  if ((statement[5][0] != '\0') && (statement[5][0] != ':')) {  // An operator was found
    fixpoint3=isfixpoint(statement[6]);
    strcpy(Areg,"invalid");
    if (index&4) loadindex(&getindex2[0]);
    if (statement[5][0] == '+') {
//      if ((fixpoint1 == 4) && (fixpoint2 == 4))
//      {

//      }
//      else
//      {
	if ((fixpoint1 == 8) && ((fixpoint2 & 8) == 8) && ((fixpoint3 & 8) == 8))
	{ //8.8=8.8+8.8
	  printf("	LDA ");
	  printfrac(statement[4]);
	  printf("	CLC \n");
	  printf("	ADC ");
	  printfrac(statement[6]);
	  printf("	STA ");
	  printfrac(statement[2]);
	  printf("	LDA ");
	  printimmed(statement[4]);
	  printf("%s\n",statement[4]);
          printf("	ADC ");
	  printimmed(statement[6]);
	  printf("%s\n",statement[6]);
	}
	else if ((fixpoint1 == 8) && ((fixpoint2 & 8) == 8) && (fixpoint3 == 4))
	{
	  printf("	LDY %s\n",statement[6]);
	  printf("	LDX ");
	  printfrac(statement[4]);
	  printf("	LDA ");
	  printimmed(statement[4]);
	  printf("%s\n",statement[4]);
	  printf("	JSR Add44to88\n");
	  printf("	STX ");
	  printfrac(statement[2]);
	}
	else if ((fixpoint1 == 4) && (fixpoint2 == 8) && ((fixpoint3 & 4) == 4))
	{
	  if (fixpoint3 == 4)
	    printf("	LDY %s\n",statement[6]);
	  else
	    printf("	LDY #%d\n",(int)(atof(statement[6])*16.0));
	  printf("	LDA %s\n",statement[4]);
	  printf("	LDX ");
	  printfrac(statement[4]);
	  printf("	JSR Add88to44\n");
	}
	else if ((fixpoint1 == 4) && (fixpoint2 == 4) && (fixpoint3 == 12))
	{
	  printf("	CLC\n");
	  printf("	LDA %s\n",statement[4]);
	  printf("	ADC #%d\n",(int)(atof(statement[6])*16.0));
	}
	else
        {
          printf("	CLC\n");
          printf("	ADC ");
          printindex(statement[6],index&4);
        }
//    }
    }
    else if (statement[5][0] == '-') {
	if ((fixpoint1 == 8) && ((fixpoint2 & 8) == 8) && ((fixpoint3 & 8) == 8))
	{ //8.8=8.8-8.8
	  printf("	LDA ");
	  printfrac(statement[4]);
	  printf("	SEC \n");
	  printf("	SBC ");
	  printfrac(statement[6]);
	  printf("	STA ");
	  printfrac(statement[2]);
	  printf("	LDA ");
	  printimmed(statement[4]);
	  printf("%s\n",statement[4]);
          printf("	SBC ");
	  printimmed(statement[6]);
	  printf("%s\n",statement[6]);
	}
	else if ((fixpoint1 == 8) && ((fixpoint2 & 8) == 8) && (fixpoint3 == 4))
	{
	  printf("	LDY %s\n",statement[6]);
	  printf("	LDX ");
	  printfrac(statement[4]);
	  printf("	LDA ");
	  printimmed(statement[4]);
	  printf("%s\n",statement[4]);
	  printf("	JSR Sub44from88\n");
	  printf("	STX ");
	  printfrac(statement[2]);
	}
	else if ((fixpoint1 == 4) && (fixpoint2 == 8) && ((fixpoint3 & 4) == 4))
	{
	  if (fixpoint3 == 4)
	    printf("	LDY %s\n",statement[6]);
	  else
	    printf("	LDY #%d\n",(int)(atof(statement[6])*16.0));
	  printf("	LDA %s\n",statement[4]);
	  printf("	LDX ");
	  printfrac(statement[4]);
	  printf("	JSR Sub88from44\n");
	}
	else if ((fixpoint1 == 4) && (fixpoint2 == 4) && (fixpoint3 == 12))
	{
	  printf("	SEC\n");
	  printf("	LDA %s\n",statement[4]);
	  printf("	SBC #%d\n",(int)(atof(statement[6])*16.0));
	}
	else
        {
          printf("	SEC\n");
          printf("	SBC ");
          printindex(statement[6],index&4);
	}
    }
    else if (statement[5][0] == '&') {
      printf("	AND ");
      printindex(statement[6],index&4);
    }
    else if (statement[5][0] == '^') {
      printf("	EOR ");
      printindex(statement[6],index&4);
    }
    else if (statement[5][0] == '|') {
      printf("	ORA ");
      printindex(statement[6],index&4);
    }
    else if (statement[5][0] == '*') {
      if ((!isimmed(statement[6])) || (!checkmul(atoi(statement[6]))))
      {
	printf("	LDY ");
        printindex(statement[6],index&4);
        if (statement[5][1] == '*') printf("	jsr mul16\n"); // general mul routine
        else printf("	jsr mul8\n");
      }
      else if (statement[5][1] == '*') mul(statement, 16);
      else mul(statement, 8);// attempt to optimize - may need to call mul anyway
      
    }
    else if (statement[5][0] == '/') {
      if ((!isimmed(statement[6])) || (!checkdiv(atoi(statement[6]))))
      {
	printf("	LDY ");
        printindex(statement[6],index&4);
        if (statement[5][1] == '/') printf("	jsr div16\n"); // general div routine
        else printf("	jsr div8\n");
      }
      else if (statement[5][1] == '/') divd(statement, 16);
      else divd(statement, 8);// attempt to optimize - may need to call divd anyway
      
    }
    else if (statement[5][0] == ':'){
      strcpy(Areg,Aregcopy); // A reg is not invalid
    } 
    else if (statement[5][0] == '('){
    // we've called a function, hopefully
      strcpy(Areg,"invalid");
      callfunction(statement);
    }
    else 
    {
      fprintf(stderr, "(%d) Error: invalid operator: %s\n",line,statement[5]);
      exit(1);
    }

  }
  else // simple assignment
  {
    // check for fixed point stuff here

    if ((fixpoint1 == 4) && (fixpoint2 == 0))
    {
      printf("  ASL\n");
      printf("  ASL\n");
      printf("  ASL\n");
      printf("  ASL\n");
    }
    else if ((fixpoint1 == 0) && (fixpoint2 == 4))
    {   
      printf("  LSR\n");
      printf("  LSR\n");
      printf("  LSR\n");
      printf("  LSR\n");
    }  
    else if ((fixpoint1 == 4) && (fixpoint2 == 8))
    {
      printf("  LDX ");
      printfrac(statement[4]);
      printf("  JSR Assign88to44\n");
    }
    else if ((fixpoint1 == 8) && (fixpoint2 == 4))
    {
      printf("  JSR Assign44to88\n");
      printf("  STX ");
      printfrac(statement[2]);
    }
    else if ((fixpoint1 == 8) && ((fixpoint2 & 8) == 8))
    {
      printf("	LDX ");
      printfrac(statement[4]);
      printf("	STX ");
      printfrac(statement[2]);
      printf("	LDA ");
      printimmed(statement[4]);
      printf("%s\n",statement[4]);
    }
    else if ((fixpoint1 == 4) && ((fixpoint2 & 4) == 4))
    {
      if (fixpoint2 == 4)
        printf("	LDA %s\n",statement[4]);
      else
        printf("	LDA #%d\n",(int)(atof(statement[4])*16));
    }
//    else if ((fixpoint1 == 0) && (fixpoint1 == 8))
//    {
//      printf("	LDA %s\n",statement[4]);
//    }
  }
    if (index&1) loadindex(&getindex0[0]);
    printf("	STA ");
    printindex(statement[2],index&1);
    free(deallocstatement);
}

void dogoto(char **statement)
{
  printf("	JMP .%s\n",statement[2]);
}

void gosub(char **statement)
{
  invalidate_Areg();
  //if (numgosubs++>3) {
   // fprintf(stderr,"Max. nested gosubs exceeded in line %s\n",statement[0]);
   // exit(1);
  //}

  printf("	JSR .%s\n",statement[2]);
}

void set(char **statement)
{
  if (!strncmp(statement[2],"smartbranching\0",14))
  {
    if (!strncmp(statement[3],"on\0",2)) smartbranching=1;
    else smartbranching=0;
  }
  else if (!strncmp(statement[2],"romsize\0",7))
  {
    set_romsize(statement[3]);
  }
  else if (!strncmp(statement[2],"optimization\0",5))
  {
    if (!strncmp(statement[3],"speed\0",5))
    {
      optimization=1;
    }
    if (!strncmp(statement[3],"size\0",4))
    {
      optimization=2;
    }
    if (!strncmp(statement[3],"none\0",4))
    {
      optimization=0;
    }
    else prerror("set: unknown parameter\n");
  }
}

void rem(char **statement)
{
  if (!strncmp(statement[2],"smartbranching\0",14))
  {
    if (!strncmp(statement[3],"on\0",2)) smartbranching=1;
    else smartbranching=0;
  }
}

void bmi(char *linenumber)
{
  removeCR(linenumber);
  if (smartbranching) {
    printf(" if ((* - .%s) < 128) && ((* - .%s) > -128)\n	bmi .%s\n",linenumber,linenumber,linenumber);
    printf(" else\n	bpl .%dskip%s\n	jmp .%s\n",branchtargetnumber,linenumber,linenumber);
    printf(".%dskip%s\n",branchtargetnumber++,linenumber);
    printf(" endif\n");
  }
  else printf("	bmi .%s\n",linenumber);
}

void bpl(char *linenumber)
{
  removeCR(linenumber);
  if (smartbranching) {
    printf(" if ((* - .%s) < 128) && ((* - .%s) > -128)\n	bpl .%s\n",linenumber,linenumber,linenumber);
    printf(" else\n	bmi .%dskip%s\n	jmp .%s\n",branchtargetnumber,linenumber,linenumber);
    printf(".%dskip%s\n",branchtargetnumber++,linenumber);
    printf(" endif\n");
  }
  else printf("	bpl .%s\n",linenumber);
}

void bne(char *linenumber)
{
  removeCR(linenumber);
  if (smartbranching) {
    printf(" if ((* - .%s) < 128) && ((* - .%s) > -128)\n	BNE .%s\n",linenumber,linenumber,linenumber);
    printf(" else\n	beq .%dskip%s\n	jmp .%s\n",branchtargetnumber,linenumber,linenumber);
    printf(".%dskip%s\n",branchtargetnumber++,linenumber);
    printf(" endif\n");
  }
  else printf("	bne .%s\n",linenumber);
}

void beq(char *linenumber)
{
  removeCR(linenumber);
  if (smartbranching) {
    printf(" if ((* - .%s) < 128) && ((* - .%s) > -128)\n	BEQ .%s\n",linenumber,linenumber,linenumber);
    printf(" else\n	bne .%dskip%s\n	jmp .%s\n",branchtargetnumber,linenumber,linenumber);
    printf(".%dskip%s\n",branchtargetnumber++,linenumber);
    printf(" endif\n");
  }
  else printf("	beq .%s\n",linenumber);
}

void bcc(char *linenumber)
{
  removeCR(linenumber);
  if (smartbranching) {
    printf(" if ((* - .%s) < 128) && ((* - .%s) > -128)\n	bcc .%s\n",linenumber,linenumber,linenumber);
    printf(" else\n	bcs .%dskip%s\n	jmp .%s\n",branchtargetnumber,linenumber,linenumber);
    printf(".%dskip%s\n",branchtargetnumber++,linenumber);
    printf(" endif\n");
  }
  else printf("	bcc .%s\n",linenumber);
}

void bcs(char *linenumber)
{
  removeCR(linenumber);
  if (smartbranching) {
    printf(" if ((* - .%s) < 128) && ((* - .%s) > -128)\n	bcs .%s\n",linenumber,linenumber,linenumber);
    printf(" else\n	bcc .%dskip%s\n	jmp .%s\n",branchtargetnumber,linenumber,linenumber);
    printf(".%dskip%s\n",branchtargetnumber++,linenumber);
    printf(" endif\n");
  }
  else printf("	bcs .%s\n",linenumber);
}

void bvc(char *linenumber)
{
  removeCR(linenumber);
  if (smartbranching) {
    printf(" if ((* - .%s) < 128) && ((* - .%s) > -128)\n	bvc .%s\n",linenumber,linenumber,linenumber);
    printf(" else\n	bvs .%dskip%s\n	jmp .%s\n",branchtargetnumber,linenumber,linenumber);
    printf(".%dskip%s\n",branchtargetnumber++,linenumber);
    printf(" endif\n");
  }
  else printf("	bvc .%s\n",linenumber);
}

void bvs(char *linenumber)
{
  removeCR(linenumber);
  if (smartbranching) {
    printf(" if ((* - .%s) < 128) && ((* - .%s) > -128)\n	bvs .%s\n",linenumber,linenumber,linenumber);
    printf(" else\n	bvc .%dskip%s\n	jmp .%s\n",branchtargetnumber,linenumber,linenumber);
    printf(".%dskip%s\n",branchtargetnumber++,linenumber);
    printf(" endif\n");
  }
  else printf("	bvs .%s\n",linenumber);
}

void drawscreen()
{
  invalidate_Areg();
  printf("	jsr drawscreen\n");
}

void prerror(char *myerror)
{
  fprintf(stderr, "(%d): %s\n", line, myerror);
}

int printimmed(char *value)
{
  int immed=isimmed(value);
  if (immed) printf("#");
  return immed;
}

int isimmed(char *value)
{
  // search queue of constants
  int i;
  // removeCR(value);
  for (i=0;i<numconstants;++i) 
    if (!strcmp(value,constants[i]))
    {
      // a constant should be treated as an immediate
      return 1;
    }

  if ((value[0] == '$') || (value[0] == '%')
     || (value[0] < (unsigned char)0x3A)) 
  {
    return 1;
  }
  else return 0;
}

int number(unsigned char value)
{
  return ((int)value)-48;
}

void removeCR(char *linenumber)  // remove trailing CR from string
{
  if ((linenumber[strlen(linenumber)-1]==(unsigned char)0x0A) ||
      (linenumber[strlen(linenumber)-1]==(unsigned char)0x0D))
      linenumber[strlen(linenumber)-1]='\0';
}

void remove_trailing_commas(char *linenumber)  // remove trailing commas from string
{
  int i;
  for (i=strlen(linenumber)-1;i>0;i--)
  {
    if ((linenumber[i] != ',') &&
      (linenumber[i] != ' ') &&
      (linenumber[i] != (unsigned char)0x0A) &&
      (linenumber[i] != (unsigned char)0x0D)) break;
    if (linenumber[i] == ',')
    {
      linenumber[i] = ' ';
      break;
    }
  }
}

void header_open(FILE *header)
{
}

void header_write(FILE *header, char *filename)
{
  int i;
  if ((header = fopen(filename, "w")) == NULL)  // open file
    {
      fprintf(stderr,"Cannot open %s for writing\n",filename);
      exit(1);
    }

  strcpy(redefined_variables[numredefvars],"; This file contains variable mapping and other information for the current project.\n");

  for (i=numredefvars;i>=0;i--)
  {
    fprintf(header, "%s\n", redefined_variables[i]);
  }
  fclose(header);
 
}
